[pve-devel] [PATCH v3 access-control 05/17] api: allow superusers to edit tfa and password settings

Oguz Bektas o.bektas at proxmox.com
Wed Apr 6 13:57:22 CEST 2022


Signed-off-by: Oguz Bektas <o.bektas at proxmox.com>
---
v2->v3:
* added `has_superuser_anywhere` helper
* only allow changing TFA or password settings of SUs if user has SU on /access endpoint


 src/PVE/API2/AccessControl.pm | 23 +++++++++++++++--------
 src/PVE/API2/TFA.pm           | 11 ++++++++++-
 src/PVE/RPCEnvironment.pm     | 15 +++++++++++++++
 3 files changed, 40 insertions(+), 9 deletions(-)

diff --git a/src/PVE/API2/TFA.pm b/src/PVE/API2/TFA.pm
index bee4dee..d7c0c4d 100644
--- a/src/PVE/API2/TFA.pm
+++ b/src/PVE/API2/TFA.pm
@@ -96,22 +96,31 @@ my $TFA_UPDATE_INFO_SCHEMA = {
 };
 
 # Only root may modify root, regular users need to specify their password.
+# Only users with SU on /access may modify other users with SU anywhere
 #
 # Returns the userid returned from `verify_username`.
 # Or ($userid, $realm) in list context.
 my sub root_permission_check : prototype($$$$) {
     my ($rpcenv, $authuser, $userid, $password) = @_;
 
+    # authuser = the user making the change
+    # userid = the user to be changed
+    raise_perm_exc() if $userid eq 'root at pam' && $authuser ne 'root at pam'; # can be only edited by itself
     ($userid, undef, my $realm) = PVE::AccessControl::verify_username($userid);
     $rpcenv->check_user_exist($userid);
 
-    raise_perm_exc() if $userid eq 'root at pam' && $authuser ne 'root at pam';
+    my $is_superuser = $rpcenv->check($authuser, "/access", ['SuperUser'], 1);
 
     # Regular users need to confirm their password to change TFA settings.
     if ($authuser ne 'root at pam') {
 	raise_param_exc({ 'password' => 'password is required to modify TFA data' })
 	    if !defined($password);
 
+	if (!$is_superuser && $authuser ne $userid) {
+	    die "only superusers can change another superuser's TFA settings!\n"
+		if $rpcenv->has_superuser_anywhere($userid);
+	}
+
 	($authuser, my $auth_username, my $auth_realm) =
 	    PVE::AccessControl::verify_username($authuser);
 
diff --git a/src/PVE/RPCEnvironment.pm b/src/PVE/RPCEnvironment.pm
index 34c8991..04c5633 100644
--- a/src/PVE/RPCEnvironment.pm
+++ b/src/PVE/RPCEnvironment.pm
@@ -479,6 +479,21 @@ sub check_api2_permissions {
     raise_perm_exc();
 }
 
+sub has_superuser_anywhere {
+    my ($self, $username) = @_;
+
+    return 1 if $username eq 'root at pam';
+
+    # get all ACL paths from config and check for SU privilege
+    my $acls = $self->{user_cfg}->{acl};
+    my @acl_paths = keys(%$acls);
+    @acl_paths = ['/'] if !@acl_paths;
+    foreach my $path (@acl_paths) {
+	return 1 if $self->check($username, $path, ['SuperUser'], 1);
+    }
+    return 0;
+}
+
 sub log_cluster_msg {
     my ($self, $pri, $user, $msg) = @_;
 
diff --git a/src/PVE/API2/AccessControl.pm b/src/PVE/API2/AccessControl.pm
index 5d78c6f..4ba9dc5 100644
--- a/src/PVE/API2/AccessControl.pm
+++ b/src/PVE/API2/AccessControl.pm
@@ -378,17 +378,24 @@ __PACKAGE__->register_method ({
 
 	$rpcenv->check_user_exist($userid);
 
+	my $is_superuser = $rpcenv->check($authuser, "/access", ['SuperUser'], 1);
+
 	if ($authuser eq 'root at pam') {
 	    # OK - root can change anything
+	} elsif ($authuser eq $userid) {
+	    $rpcenv->check_user_enabled($userid);
+	    # OK - each user can change its own password
 	} else {
-	    if ($authuser eq $userid) {
-		$rpcenv->check_user_enabled($userid);
-		# OK - each user can change its own password
-	    } else {
-		# only root may change root password
-		raise_perm_exc() if $userid eq 'root at pam';
-		# do not allow to change system user passwords
-		raise_perm_exc() if $realm eq 'pam';
+	    # only root may change root password
+	    raise_perm_exc() if $userid eq 'root at pam';
+	    # do not allow to change system user passwords
+	    raise_perm_exc() if $realm eq 'pam';
+
+	    if ($is_superuser) {
+		# OK - superusers can change any user's password except root at pam
+	    } elsif ($rpcenv->has_superuser_anywhere($userid)) {
+		die "only superusers can change another superuser's password!\n"
+		    if !$is_superuser;
 	    }
 	}
 
-- 
2.30.2






More information about the pve-devel mailing list